/****************************************************************************/
/*                                                                          */
/*                   Copyright 2003 University of Glasgow                   */
/*                                                                          */
/*                           All rights reserved.                           */
/*                                                                          */
/****************************************************************************/

/****************************************************************************/
/* FILE          : Checker.java                                             */
/* DESCRIPTION   : Typechecker for the Better Functional Language           */
/*                                                                          */
/* AUTHOR        : S.J.Gay                                                  */
/* DATE          : 28th August 2003                                         */
/*                                                                          */
/* LAST MODIFIED : S.J.Gay                                                  */
/* DATE          : 27th October 2003                                        */
/****************************************************************************/

package bfl.checker;

import java.util.*;
import java.io.*;
import bfl.node.*;
import bfl.analysis.*;
import bfl.error.*;
import bfl.types.*;
import bfl.env.*;

public class Checker extends DepthFirstAdapter {

    ErrorTable errorTable = new ErrorTable();
    Env env = new Env();

    public ErrorTable getErrorTable() {
	return errorTable;
    }

    public void outStart(Start node) {
	setOut(node, getOut(node.getPExpression()));
    }

    public void outATermExpression(ATermExpression node)
    {
        setOut(node, getOut(node.getTerm()));
    }

    public void outAPlusExpression(APlusExpression node)
    {
	Type leftType = (Type)getOut(node.getLeft());
	Type rightType = (Type)getOut(node.getRight());

	if (leftType != null) {
	    if (!(leftType instanceof IntType)) {
		errorTable.add(node.getPlus().getLine(),
			       new PlusLeftError(leftType.name()));
	    }
	};

	if (rightType != null) {
	    if (!(rightType instanceof IntType)) {
		errorTable.add(node.getPlus().getLine(),
			       new PlusRightError(rightType.name()));
	    }
	};

	if ((leftType instanceof IntType) 
	    && (rightType instanceof IntType)) {
	    setOut(node, new IntType());
	};
    }

    public void outAAndExpression(AAndExpression node)
    {
	Type leftType = (Type)getOut(node.getLeft());
	Type rightType = (Type)getOut(node.getRight());

	if (leftType != null) {
	    if (!(leftType instanceof BoolType)) {
		errorTable.add(node.getAnd().getLine(),
			       new AndLeftError(leftType.name()));
	    }
	};

	if (rightType != null) {
	    if (!(rightType instanceof BoolType)) {
		errorTable.add(node.getAnd().getLine(),
			       new AndRightError(rightType.name()));
	    }
	};

	if ((leftType instanceof BoolType) 
	    && (rightType instanceof BoolType)) {
	    setOut(node, new BoolType());
	};
    }

    public void outAEqExpression(AEqExpression node)
    {
	Type leftType = (Type)getOut(node.getLeft());
	Type rightType = (Type)getOut(node.getRight());

	if (leftType != null) {
	    if (!(leftType instanceof IntType)) {
		errorTable.add(node.getEq().getLine(),
			       new EqLeftError(leftType.name()));
	    }
	};

	if (rightType != null) {
	    if (!(rightType instanceof IntType)) {
		errorTable.add(node.getEq().getLine(),
			       new EqRightError(rightType.name()));
	    }
	};

	if ((leftType instanceof IntType) 
	    && (rightType instanceof IntType)) {
	    setOut(node, new BoolType());
	};
    }

    public void outACondExpression(ACondExpression node)
    {
	Type condType = (Type)getOut(node.getCond());
	Type thenType = (Type)getOut(node.getThenBranch());
	Type elseType = (Type)getOut(node.getElseBranch());

	if (condType != null) {
	    if (!(condType instanceof BoolType)) {
		errorTable.add(node.getIf().getLine(),
			       new IfCondError(condType.name()));
	    }
	};

	if ((thenType != null) && (elseType != null)) {
	    if (!(thenType.equal(elseType))) {
		errorTable.add(node.getIf().getLine(),
		    new IfMismatchError(thenType.name(),elseType.name()));
	    }
	    else {
		if (condType instanceof BoolType) {
		    setOut(node,thenType);
		}
	    }
	};

    }

    public void caseAFnExpression(AFnExpression node)
    {
	ATypedArg typedArg = (ATypedArg)node.getTypedArg();
	typedArg.apply(this);
	Type argType = (Type)getOut(typedArg.getTy());
	String argName = typedArg.getId().getText();
	env.openScope();
	env.put(argName,new EnvEntry(argType));
	node.getExpression().apply(this);
	Type bodyType = (Type)getOut(node.getExpression());
	if (bodyType != null) {
	    setOut(node,new FunType(argType,bodyType));
	};
	env.closeScope();
    }

    public void outAAppExpression(AAppExpression node)
    {
	Type fnType = (Type)getOut(node.getFn());
	Type actualArgType = (Type)getOut(node.getArg());

	if (fnType != null) {
	    if (!(fnType instanceof FunType)) {
		errorTable.add(-1,
			       new AppNotFunError(fnType.name()));
	    }
	    else {
		Type formalArgType = ((FunType)fnType).getArg();
		if (actualArgType != null) {
		    if (!formalArgType.equal(actualArgType)) {
			errorTable.add(-1,
				       new WrongArgTypeError
				       (formalArgType.name(),
					actualArgType.name()));
		    }
		    else {
			setOut(node, ((FunType)fnType).getRes());
		    }
		}
	    }
	}
    }

    public void caseALetExpression(ALetExpression node)
    {
	env.openScope();
	Object temp[] = node.getBinding().toArray();
	for(int i = 0; i < temp.length; i++) {
	    ((PBinding) temp[i]).apply(this);
	};
	node.getExpression().apply(this);
	setOut(node, getOut(node.getExpression()));
	env.closeScope();
    }

    public void outAExpTerm(AExpTerm node)
    {
	setOut(node, getOut(node.getExpression()));
    }

    public void outABoolLitTerm(ABoolLitTerm node)
    {
	setOut(node, new BoolType());
    }

    public void outAIntLitTerm(AIntLitTerm node)
    {
	setOut(node, new IntType());
    }

    public void outAIdTerm(AIdTerm node) {
	EnvEntry entry = env.get(node.getId().getText());
	if (entry == null) {
	    errorTable.add(node.getId().getLine(),
			   new UndefinedIdError(node.getId().getText()));
	}
	else {
	    setOut(node, entry.getType());
	};
    }

    public void outAValBinding(AValBinding node)
    {
	String id = node.getId().getText();
	Type valTy = (Type)getOut(node.getExpression());
	if (valTy != null) {
	    if (env.put(id,new EnvEntry(valTy)) != null) {
		errorTable.add(node.getId().getLine(),
			       new ValRedefinedError(node.getId().getText()));
	    }
	}
    }

    public void caseAFunBinding(AFunBinding node)
    {
	ATypedArg ta = (ATypedArg)node.getTypedArg();
	ta.apply(this);
	node.getTy().apply(this);
	Type argTy = (Type)getOut(ta.getTy());
	Type resTy = (Type)getOut(node.getTy());
	if (env.put(node.getId().getText(),new EnvEntry(new FunType(argTy,resTy))) != null) {
	    errorTable.add(node.getId().getLine(),
			   new ValRedefinedError(node.getId().getText()));
	}
	else {
	    env.openScope();
	    env.put(ta.getId().getText(),new EnvEntry(argTy));
	    node.getExpression().apply(this);
	    Type bodyTy = (Type)getOut(node.getExpression());
	    if (bodyTy != null) {
		if (!bodyTy.equal(resTy)) {
		    errorTable.add(node.getId().getLine(),
			    new WrongResultTypeError(node.getId().getText(),
							    resTy.name(),
							    bodyTy.name()));
		}
	    };
	    env.closeScope();
	}	
    }

    public void outABaseTy(ABaseTy node)
    {
	setOut(node, getOut(node.getBaseTy()));
    }

    public void outAFunTy(AFunTy node)
    {
	setOut(node, new FunType((Type)getOut(node.getArg()),
				 (Type)getOut(node.getResult())));
    }

    public void outAIntBaseTy(AIntBaseTy node) {
	setOut(node, new IntType());
    }

    public void outACompoundBaseTy(ACompoundBaseTy node)
    {
	setOut(node,getOut(node.getTy()));
    }

    public void outABoolBaseTy(ABoolBaseTy node) {
	setOut(node, new BoolType());
    }

}
